C-----------------------------------------------------------------------
      SUBROUTINE ephinit (ephname,ierr)
C
C  Open cpf ephemeris file and read into memory.
C  Convenience function using char name as input; calls ephinitu.
C
C  Input
C	ephname	- Name of cpf file (character string)
C
C  Output
C	ierr	- error code: 0= OK, -1= error
C
C  Author: R. L. Ricklefs / UT CSR
C  Date: 1 Apr, 2005
C  History:
C-----------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      CHARACTER*(*) ephname
      OPEN (2,file=ephname,err=10,status='old')
      CALL ephinitu (2,ierr)
      CLOSE (2)
      RETURN
   10 WRITE(*,*) 'Error opening ',ephname
      ierr= -1
      RETURN
      END
 
C-----------------------------------------------------------------------
      SUBROUTINE interp_pvac (jdint, seceph, outvec, backvec, 
     &aberout,aberback, corr12, corr23, fvel, ircode)
C
C  Interpolate position, velocity, aberration, and relativistic corrections
C  from cpf file arrays stored in memory.
!
!	jdint        Julian date (integer day)
!	seceph       Julian day  (seconds)
!	outvec       outv(i,1) ,...,outv(i,6): x,y,z,vx,vy,vz   outbound vector
!	backvec      backv(i,1),...,outv(i,6): x,y,z,vx,vy,vz   inbound vector
!	aberout      outbound aberration correction
!	aberback     inbound aberration correction
!	corr12       outbound relativistic correction
!	corr23       inbound relativistic correction
!       fvel         = 1 => compute velocities if not present on input
!	ircode       error code: 0=OK, 1,2=error (time out of range)
!
C  Author: R. L. Ricklefs / UT CSR
C  Date: 1 May, 2003
C  History:
C 02/09/2015 - Use the first entry where leap second flag is non-zero to correct
C              requested interpolation time. rlr.
C 06/11/2015 - Add rather than subtract the leap second. rlr.
C-----------------------------------------------------------------------

      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION aberout(3),aberback(3),outvec(6),backvec(6)
      DOUBLE PRECISION jdint,seceph
      INTEGER fvel, ircode
  
      INCLUDE  '../include/ephi.inc'

      COMMON /leap/ leapsec_itime, leapsec_value
      DOUBLE PRECISION leapsec_itime    ! =-1 if no leap second found;
                                        ! =time into cpf  of leap second
      INTEGER leapsec_value             ! value of leap second at leapsec_itime

      COMMON /deb/ debug
      LOGICAL debug
  
!     Keep the time arguments small to maintain precision.
      time=((seceph/86400.d0-jdf(1))+(jdint-jdi(1)))*86400.d0
!C        write(*,*) 'pvac seceph ',seceph,' jd: ',jdint,' time: ', tim
!C        write(*,*) 'pvac jds(1):',jds(1),' jds(nv): ',jds(nv), nv

C     When a leap second flag is non-zero, it means that a leap second has
C     occurred, but the predictions are not corrected for leap second. 
C     Therefore, If a leap second has occurred (leapsec_itime set), correct the 
C     requested intepolation time 
      if (leapsec_itime .ge. 0.0d0 .and. time .gt. leapsec_itime) then
         time = time + leapsec_value
        if (debug)
     &    write (*,*) "leap sec: itime modified by ", leapsec_value,
     &               " to ",time," after ",leapsec_itime ! print for test
      endif
  
!      CALL HERMITE(1,X,Y,Z,NMAX,NVAL,XP,YP,IRCODE)
!C        write(*,*) jds,c12
      CALL hermite (1, jds, c12, z, nv, 10, time, corr12, vz, ircode)
      IF (ircode.ne.0) then
        if (debug) then
          write(*,*) 'pvac jds(1):',jds(1),' jds(nv): ',jds(nv), nv
          write(*,*) 'time=',time
          WRITE (*,*) 'hermite error: c12'
        endif
        return
      endif
      CALL hermite (1, jds, c23, z, nv, 10, time, corr23, vz, ircode)
      IF (ircode.ne.0) then
        if (debug) WRITE (*,*) 'hermite error: c23'
          return
      endif
CC        write (*,*) 'c12,c23',corr12,corr23
      DO j=1,3
        CALL hermite (1, jds, aberoutv(1,j), z, nv, 10, time,
     &                aberout(j), vz, ircode)
        IF (ircode.ne.0) then
          if (debug) WRITE (*,*) 'hermite error: aberout',j
          return
        endif
        CALL hermite (1, jds, aberbackv(1,j), z, nv, 10, time, 
     &                aberback(j), vz, ircode)
        IF (ircode.ne.0) then
          if (debug) WRITE (*,*) 'hermite error: aberback',j
          return
        endif
      END DO
! 3 or 6 elements? Handle velocity if present
      DO j=1,6
CC        WRITE (*,*) 'pvac',j,outv(50,j),backv(50,j)
C  Outleg
        if (j .lt. 4) then
C    Position
          if (outv(1,j+3) .ne. 0.d0) then
C        ...Use both position and velocity
            CALL hermite (2, jds, outv(1,j), outv(1,j+3), nv, 10, time, 
     &        outvec(j),vz,ircode)
          else
C        ...Could create velocity or not
            CALL hermite (-fvel, jds, outv(1,j), z, nv, 10, time, 
     &        outvec(j),outvec(j+3),ircode)
          endif
C     Velocity
        else if (outv(1,j) .ne. 0.d0) then
           CALL hermite (1, jds, outv(1,j), z, nv, 10, time, 
     &       outvec(j),vz,ircode)
        endif
        IF (ircode.ne.0) then
          if (debug) WRITE (*,*) 'hermite error: outvec',j
          return
        endif
C  Backleg
        if (j .lt. 4) then
C    Position
          if (backv(1,j+3) .ne. 0.d0) then
C        ...Use both position and velocity
            CALL hermite (2, jds, backv(1,j), backv(1,j+3), nv, 10,time,
     &        backvec(j),vz,ircode)
          else
C        ...Could create velocity or not
            CALL hermite (-fvel, jds, backv(1,j), z, nv, 10, time, 
     &        backvec(j),backvec(j+3),ircode)
          endif
C     Velocity
        else if (backv(1,j) .ne. 0.d0) then
           CALL hermite (1, jds, backv(1,j), z, nv, 10, time, 
     &       backvec(j),vz,ircode)
        endif
        IF (ircode.ne.0) then
        if (debug) WRITE (*,*) 'hermite error: backvec',j
          return
        endif
        if (debug) WRITE (*,1000) j,outvec(j),backvec(j)
 1000	format ('pvac->',i2, 2(1x,f20.6))
      END DO
      if (debug) WRITE (*,*) 'pvac:',time,outvec,backvec
      RETURN
      END

C-----------------------------------------------------------------------
      SUBROUTINE interp_off (jdint, seceph, offsetout,offsetback, 
     &     ircode)
C
C  Interpolate offset from center of main body from cpf file arrays 
C  stored in memory.
!
!	jdint        Julian date (integer day)
!	seceph       Julian day  (seconds)
!	offsetout    outbound (or bounce) offset in m
!	offsetback   inbound offset in m
!	ircode       error code: 0=OK, 1,2=error (time out of range)
!
C  Author: R. L. Ricklefs / UT CSR
C  Date: 1 May, 2004
C  History:
C 02/09/2015 - Use the first entry where leap second flag is non-zero to correct
C              requested interpolation time. rlr.
C 06/11/2015 - Add rather than subtract the leap second. rlr.
C-----------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION offsetout(3), offsetback(3)
      DOUBLE PRECISION jdint,seceph
  
      INCLUDE  '../include/ephi.inc'

      COMMON /leap/ leapsec_itime, leapsec_value
      DOUBLE PRECISION leapsec_itime    ! =-1 if no leap second found;
                                        ! =time into cpf  of leap second
      INTEGER leapsec_value             ! value of leap second at leapsec_itime

      COMMON /deb/ debug
      LOGICAL debug
  
!     Keep the time arguments small to maintain precision.
      time=
     x  ((seceph/86400.d0-offsetjdf(1))+(jdint-offsetjdi(1)))*86400.d0
!C        write(*,*) 'off seceph ',seceph,' jd: ',jdint,' time: ', tim
!C        write(*,*) 'off jds(1):',jds(1),' jds(nv): ',jds(nv), nv

C     When a leap second flag is non-zero, it means that a leap second has
C     occurred, but the predictions are not corrected for leap second. 
C     Therefore, If a leap second has occurred (leapsec_itime set), correct the 
C     requested intepolation time 
      if (leapsec_itime .ge. 0.0d0 .and. time .gt. leapsec_itime) then
         time = time + leapsec_value
         if (debug) 
     &     write (*,*) "Leapsec: itime modified by ", leapsec_value,
     &                 " to ",time," after ",leapsec_itime ! print for test
      endif
  
      DO j=1,3
        CALL hermite (1, offsetjds, offsetoutv(1,j), z, nvoff, 10, 
     &                time,offsetout(j), vz, ircode)
        IF (ircode.ne.0) then
          IF (debug) WRITE (*,*) 'hermite error: offsetout',j
          RETURN
        ENDIF
        CALL hermite (1, offsetjds, offsetbackv(1,j), z, nvoff, 10, 
     &                time, offsetback(j), vz, ircode)
        IF (ircode.ne.0) then
          IF (debug) WRITE (*,*) 'hermite error: offsetback',j
          RETURN
        ENDIF
      END DO
      if (debug) WRITE (*,*) 'off:',time,offsetout, offsetback
      RETURN
      END

C-----------------------------------------------------------------------
      SUBROUTINE interp_rot (jdint, seceph, arottype, agast, arotang,
     &                       ircode)
C
C  Interpolate rotation angles (euler or north pole RA, Dec, and W)
C  from cpf file arrays stored in memory.
!
!	jdint        Julian date (integer day)
!	seceph       Julian day  (seconds)
!	arotype      Rotation type (int): 0=n/a; 1=Lunar euler; 2=ra/dec/W
!	agast        Greenwich apparent sidereal time
!	arotang      x, y, z, or ra, dec, w (3 elements, d.p. in deg)
!	ircode       error code: 0=OK, 1,2=error (time out of range)
!
C  Author: R. L. Ricklefs / UT CSR
C  Date: 1 May, 2004
C  History:
C 02/09/2015 - Use the first entry where leap second flag is non-zero to correct
C              requested interpolation time. rlr.
C 06/11/2015 - Add rather than subtract the leap second. rlr.
C-----------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION agast, arotang(3)
      DOUBLE PRECISION jdint,seceph
      INTEGER arottype
  
      INCLUDE  '../include/ephihdr.inc'
      INCLUDE  '../include/ephi.inc'

      COMMON /leap/ leapsec_itime, leapsec_value
      DOUBLE PRECISION leapsec_itime    ! =-1 if no leap second found;
                                        ! =time into cpf  of leap second
      INTEGER leapsec_value             ! value of leap second at leapsec_itime

      COMMON /deb/ debug
      LOGICAL debug
  
!     Keep the time arguments small to maintain precision.
      time=((seceph/86400.d0-rotjdf(1))+(jdint-rotjdi(1)))*86400.d0
!C        write(*,*) 'rot seceph ',seceph,' jd: ',jdint,' time: ', tim
!C        write(*,*) 'rot jds(1):',jds(1),' jds(nv): ',jds(nv), nv

C     When a leap second flag is non-zero, it means that a leap second has
C     occurred, but the predictions are not corrected for leap second. 
C     Therefore, If a leap second has occurred (leapsec_itime set), correct the 
C     requested intepolation time 
      if (leapsec_itime .ge. 0.0d0 .and. time .gt. leapsec_itime) then
         time = time + leapsec_value
        if (debug)
     &    write (*,*) "leap sec: itime modified by ", leapsec_value,
     &               " to ",time," after ",leapsec_itime ! print for test
      endif
  
      arottype= rot_type

      CALL hermite (1, rotjds, gast, z, nvrot, 10, time, agast, vz,
     &              ircode)
      IF (ircode.ne.0) then
        if (debug) then
          write(*,*) 'rot jds(1):',jds(1),' jds(nv): ',jds(nv), nv
          write(*,*) 'time=',time
          WRITE (*,*) 'hermite error: gast'
        endif
        return
      endif

      DO j=1,3
        CALL hermite (1, rotjds, rotangv(1,j), z, nvrot, 10, time,
     &                arotang(j), vz, ircode)
        IF (ircode.ne.0) then
          if (debug) WRITE (*,*) 'hermite error: rotangv',j
          return
        endif
      END DO
      if (debug) WRITE (*,*) 'rot:',time,arottype,agast,arotang
      RETURN
      END

C-----------------------------------------------------------------------
      SUBROUTINE interp_eop (jdint, seceph, x, y, dutc, ircode)
C
C  Interpolate earth orientation parameters x pole, y pole and delta UTC
C  from cpf file arrays stored in memory.
!
!	jdint        Julian date (integer day)
!	seceph       Julian day  (seconds)
!	x	     x pole position (arcsec)
!	y            y pole position (arcsec)
!	dut          ut1-utc (seconds)
!	ircode       error code: 0=OK, 1,2=error (time out of range)
!
C  Author: R. L. Ricklefs / UT CSR
C  Date: 1 May, 2004
C  History:
C 02/09/2015 - Use the first entry where leap second flag is non-zero to correct
C              requested interpolation time. rlr.
C 06/11/2015 - Add rather than subtract the leap second. rlr.
C-----------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION x, y, dutc
      DOUBLE PRECISION jdint,seceph
  
      INCLUDE  '../include/ephi.inc'

      COMMON /leap/ leapsec_itime, leapsec_value
      DOUBLE PRECISION leapsec_itime    ! =-1 if no leap second found;
                                        ! =time into cpf  of leap second
      INTEGER leapsec_value             ! value of leap second at leapsec_itime

      COMMON /deb/ debug
      LOGICAL debug
  
!     Keep the time arguments small to maintain precision.
      time=((seceph/86400.d0-eopjdf(1))+(jdint-eopjdi(1)))*86400.d0
!C        write(*,*) 'eop seceph ',seceph,' jd: ',jdint,' time: ', tim
!C        write(*,*) 'eop jds(1):',jds(1),' jds(nv): ',jds(nv), nv

C     When a leap second flag is non-zero, it means that a leap second has
C     occurred, but the predictions are not corrected for leap second. 
C     Therefore, If a leap second has occurred (leapsec_itime set), correct the 
C     requested intepolation time 
      if (leapsec_itime .ge. 0.0d0 .and. time .gt. leapsec_itime) then
         time = time + leapsec_value
         write (*,*) "itime modified by ", leapsec_value," to ",time,
     &               " after ",leapsec_itime ! print for test
      endif
 
      CALL hermite (1, eopjds, eopxy(1,1), z, nveop, 10, 
     7              time, x, vz, ircode)
6      IF (ircode.ne.0) then
        if (debug) WRITE (*,*) 'hermite error: eop-x',j
        return
      endif
      CALL hermite (1, eopjds, eopxy(1,2), z, nveop, 10, 
     &              time, y, vz, ircode)
      IF (ircode.ne.0) then
        if (debug) WRITE (*,*) 'hermite error: eop-y',j
        return
      endif
      CALL hermite (1, eopjds, eopdutc, z, nveop, 10, 
     &              time, dutc, vz, ircode)
      IF (ircode.ne.0) then
        if (debug) WRITE (*,*) 'hermite error: eop-dutc'
        return
      endif
      if (debug) WRITE (*,*) 'eop:',time,x,y,dutc
      RETURN
      END
