      SUBROUTINE HERMITE(ITYP,X,Y,Z,NMAX,NVAL,XP,YP,ZP,IRCODE)
C
C Interpolation by a polynomial using NVAL out of NMAX given data points
C
C       Input : ITYP  : +-1: Use Lagrange interpolation:
C                            Polynomial of degree NVAL-1
C                        -1: Compute first derivative ZP, too
C
C                         2: Use Hermite interpolation
C                            including given first derivatives:
C                            Polynomial of degree 2*NVAL-1
C
C               NVAL  : number of points to use for interpolation
C               NMAX  : number of given points in list
C               X(I)  : arguments of given values (I=1,...,NMAX)
C               Y(I)  : functional values Y=f(X)
C               Z(I)  : derivatives       Z=f(X) (ITYP=2)
C               XP    : interpolation argument
C
C       Output: YP    : interpolated value at XP
C               ZP    : first derivative of YP (ITYP=-1 only)
C               IRCODE: return code
C                       0 = ok
C                       1 = interpolation not in center interval
C                       2 = argument out of range, value returned for
c                           minimum or maximum possible argument
c
C  The function selects the NVAL values to be used for interpolation such
C  that the interpolated data point is located in the center interval.
C  (Works best for NVAL = even number, of course).
C
c  A warning code is generated if the data point is too close to the
c  edges or outside of the given arguments.
C
C  Author:   W. Gurtner
C            Astronomical Institute
C            University of Bern
C
C  Created:  June 2002
C
C  Modified: 25-JUL-2005 : Check boundaries
C
C
      IMPLICIT REAL*8 (A-H,O-Z)
C
      DIMENSION X(*),Y(*),Z(*)
C
      IRCODE=0
      YP=0.D0
      ZP=0.D0
      N=NVAL-1
      IF(XP.LT.X(1)) THEN
        WRITE(*,*) 'SR HERMITE: Argument XP set to minimum'
        IRCODE=2
        XP=X(1)
      END IF
      IF(XP.GT.X(NMAX)) THEN
        WRITE(*,*) 'SR HERMITE: Argument XP set to maximum'
        IRCODE=2
        XP=X(NMAX)
      END IF
C
C  LOOK FOR GIVEN VALUE IMMEDIATELY PRECEEDING INTERPOLATION ARGUMENT
C
      DO 210 I=1,NMAX
        IF(X(I).GE.XP) GOTO 220
210   CONTINUE
C
C  START INDEX (FIRST VALUE TO BE USED FOR INTERPOLATION) IN VECTORS X,Y,Z
C
220   I0=I-(N+1)/2
      IF(I0.LE.0) THEN
        WRITE(*,*) 'SR HERMITE: Interpolation not in center interval'
        I0=1
        IRCODE=1
      END IF
      IF(I0+N.GT.NMAX) THEN
        WRITE(*,*) 'SR HERMITE: Interpolation not in center interval'
        I0=NMAX-N
        IRCODE=1
      END IF
C
C  LAGRANGE FORMULA FOR POLYNOMIAL INTERPOLATION
C  *********************************************
C
      IF(IABS(ITYP).EQ.1) THEN
C
        DO 110 I=0,N
C
          PJ=1.D0
          DO 130 J=0,N
            IF(J.NE.I) PJ=PJ*(XP-X(J+I0))/(X(I+I0)-X(J+I0))
130       CONTINUE
C
          YP=YP+Y(I+I0)*PJ
C
110     CONTINUE
C
C  COMPUTE DERIVATIVE OF THE LAGRANGE POLYNOMIAL
C
        IF(ITYP.EQ.-1) THEN
          DO 310 I=0,N
            SK=0.D0
            DO 320 K=0,N
              IF(K.NE.I) THEN
                PJ=1.D0
                DO 330 J=0,N
                  IF(J.NE.I.AND.J.NE.K) THEN
                    PJ=PJ*(XP-X(J+I0))/(X(I+I0)-X(J+I0))
                  END IF
330             CONTINUE
                SK=SK+PJ/(X(I+I0)-X(K+I0))
              END IF
320         CONTINUE
            ZP=ZP+Y(I+I0)*SK
310       CONTINUE
        END IF
C
C  HERMITE INTERPOLATION (ADDITIONAL USE OF THE DERIVATIVES)
C  *********************************************************
C
      ELSEIF(ITYP.EQ.2) THEN
C
        DO 10 I=0,N
C
          SK=0.D0
          DO 20 K=0,N
            IF(K.NE.I) SK=SK+1.D0/(X(I+I0)-X(K+I0))
20        CONTINUE
          VI=1.D0-2.D0*(XP-X(I+I0))*SK
          UI=XP-X(I+I0)
C
          PJ=1.D0
          DO 30 J=0,N
            IF(J.NE.I) PJ=PJ*(XP-X(J+I0))/(X(I+I0)-X(J+I0))
30        CONTINUE
C
          YP=YP+(Y(I+I0)*VI+Z(I+I0)*UI)*PJ*PJ
C
10      CONTINUE
      END IF
C
      RETURN
      END
