#if ! defined( _ELEMENT_H_ )
#define _ELEMENT_H_

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "YourTimeDateClass.h"
#include <string>
#include <vector>
#include <sstream>

//-----------------------------------------------------------------------------
// Template function to convert a std::stringstream to a basic type.
// Success indicated by the ok flag.
//-----------------------------------------------------------------------------/
template <class T>
inline T convert( std::istringstream& iss, bool& ok )
{
    T result; 
    iss.clear();
    if( !(iss >> result ) )
        ok = false;
    return result;
}

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// The Element class is the set of data associated with orbital elements.
// A sub-set of these data will be common to all types of elements, and
// the remaining are described by a number of descended classes.
//
// Remarks:
//-----------------------------------------------------------------------------
class Element
{

protected :

    //  Constructor  
    Element()
    {
        m_objectName = ""; 
        m_supplierId = "";   
        m_elemType   = Element::Type::unknown;   
        m_targetId   = 0;
    };

public :

    //  Destructor.
    virtual ~Element(){};

    //  Enumeration of known element types including IRVs,
    // tuned IRVs, TLEs, Keplarian elements, state vector, etc. for earth orbit
    // objects, and lunar elements for lunar targets, elements for stars etc.
    // Terrestrial targets may be considered to have "fixed" elements which may be
    // a set of static coordinates.. 
    enum Type
    {
        unknown = 0,        // Unknown element type
        irv,                // Inter Range Vector (Tuned)
        tle,                // Three Line Element (e.g. from Norad catalog)
        cpf,                // ILRS Combined Prediction Format
        keplarian,          // Keplarian element
        nTypes              // Must be last entry in enum structure
    };

    //------------
    // Set methods
    //------------
    //  Method to set the (short) name of the associated object
    void objectNameIs(const std::string& name){m_objectName = name;};

    //  Method to set the unique identifier of the associated object
    void targetIdIs(long identifier){m_targetId = identifier;};  

    //  Method to set the identifier of supplier of element data
    void supplierIs(const std::string& supplier){m_supplierId = supplier;};  

    //  Method to set the time of element creation
    void createdIs( YourTimeDateClass date){m_created = date;}; 

    //  Method to set the reference time of an element
    void epochIs( YourTimeDateClass date){m_epoch = date;};  

    //  Method to set the element type
    void elementTypeIs(Type type){m_elemType = type;}; 

    //---------------
    // Access methods
    //---------------

    //  Method to access the (short) name of the associated object 
    std::string objectName() const {return m_objectName;};

    //  Method to access the unique identifier of the associated object 
    long targetId() const {return m_targetId;}; 

    //  Method to access the supplier of element data
    std::string supplier() const {return m_supplierId;}; 

    //  Method to access the Time of element creation
    YourTimeDateClass created() const {return m_created;}; 

    //  Method to access the Reference time of an element
    YourTimeDateClass epoch() const {return m_epoch;}; 

    //  Method to access the element type
    Type elementType() const {return m_elemType;}; 

    // Method to determine the element supplier, if possible
    bool findSupplier( const std::string &text );

protected :

    std::string        m_objectName; // Name of the associated object 
    std::string        m_supplierId; // Id of supplier of element data
    long               m_targetId;   // Unique integer id of a target object
    YourTimeDateClass  m_created;    // Time of element creation
    YourTimeDateClass  m_epoch;      // Reference time of an element
    Type               m_elemType;   // element type

};  // class Element

#endif

 

