#include "Track.h"
#include "YourException.h"

/******************************************************************************
 Constructor
******************************************************************************/
LunarTrack::LunarTrack( YourTimeDateClass     startPredictions,
                             YourTimeDateClass     stopPredictions,
                             YourTimeIntervalClass stepSize ) : OrbitalTrack( startPredictions, stopPredictions, stepSize )
{
    m_trackType = Track::Type::lunar;
}

/******************************************************************************
 Destructor
******************************************************************************/
LunarTrack::~LunarTrack( )
{
}

/******************************************************************************
Method to add a new node i.e. a set of prediction data at a given point to the
time series of such points. It also sets up a number of arrays for data
associated with each prediction segment. One segment corresponds to predictions
generated from a given set of predictions. i.e. from a CPF file.
These arrays include for each segment
1. start and stop time - time segments may overlap
2. start and stop index 
3. if we have return leg predictions, velocity, etc.
******************************************************************************/
void LunarTrack::appendNode( YourTimeDateClass epoch,
                                  bool newSegment, bool biDirectional, bool cCorrApplied,
                                  const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                                  YourTimeIntervalClass ut1_utc, YourTimeIntervalClass gast,
                                  const YourMatrix<double>& targetOffsets )
{
    OrbitalTrack::appendNode(epoch, newSegment, biDirectional, cCorrApplied, posVelCor, xPolar, yPolar, ut1_utc);

    // If these data are the first of a new segment, generated from a fresh
    // element, then update the time and index arrays. Otherwise, it is
    // a continuation of current segment.
    if( newSegment )
    {
        m_rotationType.push_back( long( targetOffsets[1][3]) );
    }
    m_aGast.push_back(  gast );
    m_aXOffsetTx.push_back( targetOffsets[0][0] );
    m_aYOffsetTx.push_back( targetOffsets[1][0] );
    m_aZOffsetTx.push_back( targetOffsets[2][0] );
    m_aXOffsetRx.push_back( targetOffsets[0][1] );
    m_aYOffsetRx.push_back( targetOffsets[1][1] );
    m_aZOffsetRx.push_back( targetOffsets[2][1] );
    m_aRotAngle1.push_back( targetOffsets[0][2] );
    m_aRotAngle2.push_back( targetOffsets[1][2] );
    m_aRotAngle3.push_back( targetOffsets[2][2] );
    return;
}


/******************************************************************************
Compute the lunar libration rotation matrix using the information
contained in the libration (euler) angles from the jpl ephemeris.
Note: the derivative positions are not populated.
Input euler libration angles (radians).
Author: R. L. Ricklefs / UT CSR
******************************************************************************/
Matrix<double> LunarTrack::librationMatrix( SiteHandle& site )
{
    YourMatrix<double> m(3,3);

    double c1 = cos( site.m_rotAngle[0] );
    double c2 = cos( site.m_rotAngle[1] );
    double c3 = cos( site.m_rotAngle[2] );
    double s1 = sin( site.m_rotAngle[0] );
    double s2 = sin( site.m_rotAngle[1] );
    double s3 = sin( site.m_rotAngle[2] );

    m(1,1) = c3* c1 - c2* s1* s3;
    m(2,1) = c3* s1 + c2* c1* s3;
    m(3,1) = s3* s2;
    m(1,2) =-s3* c1 - c2* s1* c3;
    m(2,2) =-s3* s1 + c2* c1* c3;
    m(3,2) = c3* s2;
    m(1,3) = s2* s1;
    m(2,3) =-s2* c1;
    m(3,3) = c2;
    return m;
}


/******************************************************************************
 Routine to interpolate required data at a given specific time.
******************************************************************************/
bool LunarTrack::interpolateData(SiteHandle& site, YourTimeDateClass epoch ) 
{
    if( !OrbitalTrack::interpolateData( site, epoch ) )
        return false;

    site.m_gast            = lagrangianInterpolation( m_aGast );
    site.m_targetOffset[0] = lagrangianInterpolation( m_aXOffsetTx );
    site.m_targetOffset[1] = lagrangianInterpolation( m_aYOffsetTx );
    site.m_targetOffset[2] = lagrangianInterpolation( m_aZOffsetTx );
    site.m_rotAngle[0]     = lagrangianInterpolation( m_aRotAngle1 );
    site.m_rotAngle[1]     = lagrangianInterpolation( m_aRotAngle2 );
    site.m_rotAngle[2]     = lagrangianInterpolation( m_aRotAngle3 );

    if( m_bestPrecision )
    {
        site.m_targetOffset[3] = lagrangianInterpolation( m_aXOffsetRx );
        site.m_targetOffset[4] = lagrangianInterpolation( m_aYOffsetRx );
        site.m_targetOffset[5] = lagrangianInterpolation( m_aZOffsetRx );
    }
    return true;
}

/******************************************************************************
 Routine to interpolate required data at a given specific time.
******************************************************************************/
bool LunarTrack::extractNodeData(SiteHandle& site, long index ) 
{

    if( index < 0 || index >= long(m_aX.size()) )
        return false;

    if( !OrbitalTrack::extractNodeData( site, index ) )
        return false;

    site.m_gast            =  m_aGast[index];
    site.m_targetOffset[0] =  m_aXOffsetTx[index];
    site.m_targetOffset[1] =  m_aYOffsetTx[index];
    site.m_targetOffset[2] =  m_aZOffsetTx[index];
    site.m_rotAngle[0]     =  m_aRotAngle1[index];
    site.m_rotAngle[1]     =  m_aRotAngle2[index];
    site.m_rotAngle[2]     =  m_aRotAngle3[index];

    if( m_bestPrecision )
    {
        site.m_targetOffset[3] =  m_aXOffsetRx[index];
        site.m_targetOffset[4] =  m_aYOffsetRx[index];
        site.m_targetOffset[5] =  m_aZOffsetRx[index];
    }
    return true;
}

/******************************************************************************
 Routine to calculate required angle coordinate given a specific time within the
 available range of data. 
******************************************************************************/
bool LunarTrack::processDataPoint( SiteHandle& site, YourTimeDateClass epoch, long index )
{
    // Convert rectangular coordinates to topocentric coordinates for the given site.
    if( !convertITRFToTopocentric( site, epoch, index ) )
        return false;

    // Add relativistic corrections and delays and em geodesic curvature
    site.m_targetRange += site.m_relCor[0] + site.m_relCor[1] + YourTimeDateClass::picoSeconds(270,0);
    site.m_targetRange *= relativisticScale( epoch, site.m_posSpherical[4], site.m_posSpherical[5] );
    site.m_targetRange += m_targetResponseDelay;

    // If the target is below zero degree horizon, then return false;
    normalizeAngles( &site.m_previousAz, &site.m_targetPosSpher[1], &site.m_targetPosSpher[2] );

    updateTargetCharacteristics( site );
    site.m_targetNearMoon = true;

    if( site.targetAboveHorizon() )
    {
        // Determine if the point between moon and observer is sum illuminated,
        // and if so, set to standard moon magnitude.
        double xm = site.m_targetPosRect[0];
        double ym = site.m_targetPosRect[1];
        double zm = site.m_targetPosRect[2];
        double mjd = YourTimeDateClass::modJulianDate( epoch );
        site.m_targetIlluminated = SunIlluminated(  mjd, xm, ym, zm, site.m_geodeticLatitude, site.m_longitude );
        
        // Calculate lunar phase & magnitude
        double phase = (mjd - MJD2000 - 5.1);
        phase -= floor(phase);
        if( phase < 0 )
            phase++;
        if( phase > 0.1 )
            site.m_targetMagnitude = -12 - 2.5*log10(phase);
    }

    return true;
}
 
/******************************************************************************
******************************************************************************/
void LunarTrack::checkOffsets(bool inBound, SiteHandle& site, 
                                   YourMatrix<double>& pos,
                                   YourMatrix<double>& vel,
                                   YourTimeIntervalClass& range )
{
    // Use target offset data if available
    if( !m_aXOffsetTx.empty() )
    {
        if( m_rotationType[site.m_segment] == 1 )
        { 
            // Rotation type is for lunar libration. Apply to the (lunar) target
            YourMatrix<double> offsets (3, 1, &site.m_targetOffset[inBound?3:0] );
            offsets *= librationMatrix( site );

            // Rotate from inertial to (earth) body fixed
            double gast = TWO_PI * site.m_gast.days();
            if( inBound )
            {
                gast *= -1;
                offsets *= -1;
            }
            pos = rotationMatrix3<double>(-gast) * ( rotationMatrix3<double>( gast) * pos + offsets );
            vel = rotationMatrix3<double>(-gast) * ( rotationMatrix3<double>( gast) * vel );
            range =  YourTimeDateClass::seconds( 1.0e6 * pos.norm()/SpeedOfLight );
            return;
        }
        else if( m_rotationType[site.m_segment] == 2 )
        {
            // TBD
            yourTHROW( "Other rotation types not defined.");
        }
    }

    return;
}

