#pragma warning( disable : 4786 )

#if ! defined( _SITEHANDLE_H_ )
#define _SITEHANDLE_H_

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <string>
#include <vector>
#include "yourTimeClass.h"


//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// This class supports the association of prediction data stored in a Track
// object with a given site. For a given SiteHandle object, site details must
// be defined, and the object will store prediction data pertaining to that site
// for a given epoch. 
//-----------------------------------------------------------------------------
class SiteHandle
{
public:
    // Constructor
    // Arguments: name - site name.
    //            geodeticLatitude - geodetic latitude in degrees.
    //            longitude - geodetic longitude in degrees.
    //            height    -  height above the ellipsoid in meters.
    //            x, y, z - Coordinates in a local reference frame may be defined if required. 
    SiteHandle( const std::string& name, double geodeticLatitude, double longitude, double height, double horizon = 20,
                double x = 0, double y = 0, double z = 0);

    // Destructor
    virtual ~SiteHandle( );

    //============
    // Set methods
    //============
    // Method to set the observing horizon in degrees.
    void setHorizonDegrees(double horizon){ m_horizon = RAD_PER_DEG * horizon; };

    // Method to set the identifier of the site. 
    void setIdentifier( const std::string& id ){ m_identifier = id;};
    
    //===============
    // Access methods
    //===============
    // Method to access site name.
    std::string name() const { return m_name;}

    // Method to access the epoch of the current data point.
    YourTimeDateClass epoch() const { return m_epoch;} 

    // Method to access site id.
    std::string identifier() const { return m_identifier;}

    // Method to access elevation in radians.
    double elevationRadian()    const { return m_targetMeanSpher[2];}
    double elevationRadianIn()  const { return m_targetPosSpher[5];}
    double elevationRadianOut() const { return m_targetPosSpher[2];}

    // Method to access elevation in degrees.
    double elevationDegree()    const { return m_targetMeanSpher[2] * DEG_PER_RAD;}
    double elevationDegreeIn()  const { return m_targetPosSpher[5]  * DEG_PER_RAD;}
    double elevationDegreeOut() const { return m_targetPosSpher[2]  * DEG_PER_RAD;}

    // Method to access azimuth in radians.
    double azimuthRadian()    const { return m_targetMeanSpher[1];}
    double azimuthRadianIn()  const { return m_targetPosSpher[4];}
    double azimuthRadianOut() const { return m_targetPosSpher[1];}

    // Method to access azimuth in degrees.
    double azimuthDegree()    const { return m_targetMeanSpher[1] * DEG_PER_RAD;}
    double azimuthDegreeIn()  const { return m_targetPosSpher[4]  * DEG_PER_RAD;}
    double azimuthDegreeOut() const { return m_targetPosSpher[1]  * DEG_PER_RAD;}

    // Method to access range in megameters.
    double rangeMm()    const { return m_targetMeanSpher[0];} 
    double rangeInMm()  const { return m_targetPosSpher[3];} 
    double rangeOutMm() const { return m_targetPosSpher[0];} 

    // Method to access two way range in light propagation.
    YourTimeIntervalClass range() const { return m_targetRange;} 

    // Method to access elevation velocity in radians per sec.
    double elevationVelocityRps()    const { return m_targetMeanSpher[5];}
    double elevationVelocityInRps()  const { return m_targetVelSpher[5];}
    double elevationVelocityOutRps() const { return m_targetVelSpher[2];}

    // Method to access elevation velocity in degrees per sec.
    double elevationVelocityDps()    const { return m_targetMeanSpher[5] * DEG_PER_RAD;}
    double elevationVelocityInDps()  const { return m_targetVelSpher[5]  * DEG_PER_RAD;}
    double elevationVelocityOutDps() const { return m_targetVelSpher[2]  * DEG_PER_RAD;}

    // Method to access azimuth velocity in radians per sec.
    double azimuthVelocityRps()    const { return m_targetMeanSpher[4];}
    double azimuthVelocityInRps()  const { return m_targetVelSpher[4];}
    double azimuthVelocityOutRps() const { return m_targetVelSpher[1];}

    // Method to access azimuth velocity in degrees per sec.
    double azimuthVelocityDps()    const { return m_targetMeanSpher[4] * DEG_PER_RAD;}
    double azimuthVelocityInDps()  const { return m_targetVelSpher[4]  * DEG_PER_RAD;}
    double azimuthVelocityOutDps() const { return m_targetVelSpher[1]  * DEG_PER_RAD;}

    // Method to access range velocity in megameters per sec.
    double rangeVelocityMmps()    const { return m_targetMeanSpher[3];}
    double rangeVelocityInMmps()  const { return m_targetVelSpher[3];}
    double rangeVelocityOutMmps() const { return m_targetVelSpher[0];}

    // Method to access magnitude of target velocity projected on the
    // az-el plane in radians per sec.
    double velocityAzElRps() const { return m_targetVelocityAzEl;}

    // Method to access magnitude of target velocity projected on the
    // az-el plane in degrees per sec.
    double velocityAzElDps() const { return m_targetVelocityAzEl * DEG_PER_RAD;}

    // Method to access target right ascension in degrees.
    double targetRAHour() const { return m_targetRA;}

    // Method to access target declination in degrees.
    double targetDecDegree() const { return m_targetDec;}

    // Method to access target ITRF position in Mm.
    double xTargetInMm()  const { return m_targetPosRect[3];}
    double yTargetInMm()  const { return m_targetPosRect[4];}
    double zTargetInMm()  const { return m_targetPosRect[5];}
    double xTargetOutMm() const { return m_targetPosRect[0];}
    double yTargetOutMm() const { return m_targetPosRect[1];}
    double zTargetOutMm() const { return m_targetPosRect[2];}

    // Method to access illuminated flag. The observing
    // site must be in nightimte and the target reflecting sun light or
    // generating own light.
    bool isIlluminated() const { return m_targetIlluminated;}

    // Method to access sun proximity flag.
    bool isNearSun() const { return m_targetNearSun;}

    // Method to access moon proximity flag.
    bool isNearMoon() const { return m_targetNearMoon;}

    // Method to access magnitude of target illumination.
    double visualMagnitude() const { return m_targetMagnitude;}

    // Method to determine if target is above the local (set) horizon.
    bool targetAboveHorizon() const { return m_targetPosSpher[2] > m_horizon;};

    // Method to determine if target is above the zero horizon.
    bool targetRisen() const { return m_targetPosSpher[2] > 0;};

private:

    friend class Track;
    friend class OrbitalTrack;
    friend class LunarTrack;
    friend class TransponderTrack;
    std::string m_name;                     // site name
    std::string m_identifier;               // site identifier
    double m_height;                        // site height above ellipsoid (Mm)
    double m_horizon;                       // Operating horizon (rads)
    double m_geodeticLatitude;              // site geodetic latitude (rads)
    double m_longitude;                     // site longitude east (rads)
    double m_posITRF[3];                    // Site XYX ITRF positions (Mm)
    double m_posLocal[3];                   // Site local XYZ coordinates
    double m_posSpherical[3];               // Site spherical (radius,long,lat) coordinates
    double m_targetPosRect[6];              // Target XYZ ITRF positions outBound and inbound
    double m_targetVelRect[6];              // Target XYZ ITRF velocities, outBound and Inbound
    double m_targetPosSpher[6];             // Target spherical positions wrt the site, outBound, Inbound
    double m_targetVelSpher[6];             // Target spherical velocities wrt the site, outBound, Inbound
    double m_targetMeanSpher[6];            // Average target spherical position, velocity wrt the site
    double m_aberCor[6];                    // Stellar aberrations, relativistic correction outBound and inbound
    double m_targetOffset[6];               // Target offsets, outBound and inbound
    double m_targetRA;                      // target right ascension
    double m_targetDec;                     // target declination
    double m_rotAngle[3];                   // Target rotation angles
    double m_targetVelocityAzEl;            // Projected Velocity 
    double m_transponderCor;                // Transponder oscillator relativity correction
    double m_xPolar;                        // Polar x-motion
    double m_yPolar;                        // Polar y-motion
    YourTimeIntervalClass m_targetRange;   // Total target range as two-way time of flight. 
    YourTimeIntervalClass m_relCor[2];     // Relativistic corrections, outbound and inbound 
    YourTimeIntervalClass m_gast;          // Greenwich Apparent Sidereal Time 
    YourTimeIntervalClass m_ut1_utc;       // UT1 - UTC 
    YourTimeDateClass     m_epoch;         // epoch of these data 
    bool   m_targetIlluminated;             // illuminated by sun?
    bool   m_targetNearSun;                 // near the sun?
    bool   m_targetNearMoon;                // near the moon?
    double m_targetMagnitude;               // visual magnitude
    double m_previousAz;                    // Previous azimuth value
    long   m_index;                         // Current prediction index
    long   m_segment;                       // Current segment
    long   m_previousIndex;                 // Last prediction index used

};	   // class SiteHandle
#endif // _SITEHANDLE_H_
