#include "Track.h"
#include "YourException.h"


/******************************************************************************
 Constructor
******************************************************************************/
Track::Track(  YourTimeDateClass     startPredictions,
               YourTimeDateClass     stopPredictions,
               YourTimeIntervalClass stepSize )
{
    m_trackType = Track::Type::undefined;
    yourASSERTION( YourTimeDateClass::seconds(stepSize) > 0 );
    yourASSERTION( stopPredictions > startPredictions );
    m_startEpoch = m_startPredicts = m_periodStart = startPredictions;
    m_stopEpoch  = m_stopPredicts  = m_periodStop  = stopPredictions;
        
    m_stepSize      = stepSize;
    m_bestPrecision = true;

    m_segmentStartTime.clear();
    m_segmentStopTime.clear();
    m_segmentStartIndex.clear();
    m_segmentStopIndex.clear();
    m_aX.clear();
    m_aY.clear();
    m_aZ.clear();
    m_aU.clear();
    m_aV.clear();
    m_aW.clear();
    m_maxIndex     = -1;
    m_numOfNodes   = 0;
    m_sunHalo      = 18.0;
    m_moonHalo     = 10.0;
    m_numInterpMax = 8;
    m_numInterp    = 0;
    m_lagCoeffs.clear();
    m_lagIndex.clear();
    m_lagCoeffs.resize(m_numInterpMax);
    m_lagIndex.resize(m_numInterpMax);
    m_ignoreHorizon = false;
}

/******************************************************************************
 Destructor
******************************************************************************/
Track::~Track( )
{
}

/******************************************************************************
Method to set the number of points used by the Lagrangian interpolator.
******************************************************************************/
void Track::setInterpolatorWidth(long number)
{
    m_numInterpMax = number;
    if( number < 2 )
        m_numInterpMax = 2;
    m_lagCoeffs.clear();
    m_lagIndex.clear();
    m_lagCoeffs.resize(m_numInterpMax);
    m_lagIndex.resize(m_numInterpMax);

    return;
};
/******************************************************************************
Method to find the epoch associated with an index. Refer to the getIndex
method for a full explanation.
******************************************************************************/
YourTimeDateClass Track::getEpoch( long index, long *segment )
{
    YourTimeDateClass epoch;
    epoch.setDate(0,0,0,0,0,0,0,0);
    if( segment ) *segment = -1;
    if( index >= 0 && index <= m_maxIndex && !m_segmentStartTime.empty() )
    {
        long s = 0;
        do
        {
            if( index <= m_segmentStopIndex[s] )
            {
                unsigned long i = index - m_segmentStartIndex[s];
                epoch = m_segmentStartTime[s] + i * m_stepSize;
                if( segment ) *segment = s;
                return epoch;
            }
            s++;
        }while( s < long ( m_segmentStartTime.size() ) );
    }
    return epoch;
}


/******************************************************************************
Method to find index to prediction data given a time epoch. The prediction
data should have overlapped data to allow continuation of tracking across
boundaries between segments of orbit generated from different elements. A
list of data for the different segments is maintained for this process and
includes start end end times of each segment and the index of the first node
in each segment. 

To select the appropriate segment, use the following rules.
1. Where there is no overlap, there is no choice but to use the single 
available segment.
2. If the previous request did not generate a track above the horizon
(last index = -1) then use the later segment.
3. If the previous request did generate a track above the horizon, then
use the last index to select the same segment as before.

  In the following, t1, t2, t3, and t4 refer to the following markers;

                           t3
  --------------------------                t2
  (t3)             -------------i------------- 
                   t1                   ----------------------           (t4)
                                        t4             ----------------------

******************************************************************************/
long Track::getIndex( YourTimeDateClass epoch, long *lastIndex, long *segment )
{
    long index = -1;
    *segment = -1;

    // If a set of nodes are not available, then return -1.
    if( m_segmentStartTime.empty() )
        return index;

    unsigned long n = (unsigned long) m_segmentStartTime.size() - 1;
    yourASSERTION( n >= 0 );

    // Check if the requested time is is within the prediction period
    if( epoch < m_segmentStartTime[0] || epoch > m_segmentStopTime[n])
        return index;
    
    // Search the segment list
    YourTimeDateClass t1, t2, t3, t4;
    for(unsigned long i = 0; i <= n; i++ )
    {
        t1 = m_segmentStartTime[i];
        t2 = m_segmentStopTime[i];
        if( i == 0 ) t3 = t1;
        else t3 = m_segmentStopTime[i-1];

        // Extend t4 a lttle, so that the test, epoch < t4, succeeds and 
        // the last point is included.
        if( i == n ) t4 = t2 + YourTimeDateClass::seconds(1); 
        else t4 = m_segmentStartTime[i+1];

        if( epoch >= t3 && epoch < t4 )
        {
            // Check there is no gap
            if( epoch > t2 )
                yourTHROW(" Illegal gap in prediction data");

            index = m_segmentStartIndex[i] + 
                   (long) floor(YourTimeDateClass::seconds( epoch - t1 )/YourTimeDateClass::seconds( m_stepSize ));
            *lastIndex = index;
            *segment = i;
            return index;
        }
        if( epoch >= t4 && epoch < t2 )
        {
            if( *lastIndex < 0 || *lastIndex >= m_segmentStartIndex[i+1] )
            {
                index = m_segmentStartIndex[i+1] + 
                       (long) floor(YourTimeDateClass::seconds( epoch - t4 )/YourTimeDateClass::seconds( m_stepSize ));
                *lastIndex = index;
                *segment = i+1;
                return index;
            }
            index = m_segmentStartIndex[i] + 
                   (long) floor(YourTimeDateClass::seconds( epoch - t1 )/YourTimeDateClass::seconds( m_stepSize ));
            *lastIndex = index;
            *segment = i;
            return index;
        }
    }
    return -1;
}


/******************************************************************************
 This routine calculates the coefficients for interpolation using the 
 Lagrangian equation. e.g. for a 8 point interpolator, use

    ( t-t1)( t-t2)..(t-t7)       ( t-t0)( t-t2)..( t-t7)           ( t-t0)( t-t1)..( t-t6)
x = ----------------------- x0 + ----------------------- x1 + ...+ ----------------------- x7
    (t0-t1)(t0-t2)..(t0-t7)      (t1-t0)(t1-t2)..(t1-t7)           (t7-t0)(t7-t1)..(t7-t6)


 Reference: Eq (3.1.1), P. 102, Numerical Recipes in Fortran, William Press, et al.
******************************************************************************/
void Track::generateLagrangianCoefficients( long index, long segment,
                                            YourTimeDateClass epoch)
{
    // This routine only applicble if used with a series of nodes.
    if( m_segmentStartTime.empty() )
        yourTHROW("No node data available for interpolation");

    m_numInterp = m_numInterpMax;
    double interval = YourTimeDateClass::seconds(m_stepSize);

    // Identify if this is the last segment
    bool bLastSegment = ( segment == (m_segmentStartIndex.size() - 1));

    // Find index of starting interpolation node
    long index0 = index - (m_numInterp/2 - 1);

    // If the starting node precedes the first index in the segment then
    // set it to the first index (i.e. remain in the same segment).
    if( index0 < m_segmentStartIndex[segment] )
        index0 = m_segmentStartIndex[segment];

    // If the last node exceeds the last index in the segment, and
    // this is the last segment, then shift starting node earlier.
    if( (index0 + m_numInterp) > m_segmentStopIndex[segment] && bLastSegment )
    {
        long n = m_segmentStopIndex[segment] - m_segmentStartIndex[segment] + 1;
        if( n < 2 ) yourTHROW("No data for Lagrangian interpolation.");
        if( n < m_numInterp ) m_numInterp = n;
        index0 = m_segmentStopIndex[segment] - m_numInterp;
    }

    // Determine the time difference between the end of this and the start of the next
    // segment. Use this to determine if there is an overlap, and if so, how entries
    // to skip over to get to the next time mark. 
    long skip = 0;
    double offsetTime = 0;
    if( !bLastSegment )
    {
        double delta = YourTimeDateClass::seconds(m_segmentStopTime[segment] - m_segmentStartTime[segment+1] );

        // Add 20% of interval to ensure that time intervals are not too small.
        skip = (long)( 1 + floor(0.2 + delta/interval) );
        if( skip < 0 ) skip = 0;
        offsetTime = skip * interval - delta;
    }

    // Set up an array of indices and times, including a jump to the next
    // segment if necessary.
    std::vector<double> aTime( m_numInterp );
    aTime[0]  = 0;
    m_lagIndex[0] = index0;
    bool bJump = true;
    for (long i = 1; i < m_numInterp; i++)
    {
        if( i + index0 > m_segmentStopIndex[segment] && bJump)
        {
            m_lagIndex[i] =  m_lagIndex[i-1] + 1 + skip;
            aTime[i] = aTime[i-1] + offsetTime;
            bJump = false;
        }
        else
        {
            m_lagIndex[i] = m_lagIndex[i-1] + 1;
            aTime[i] =  aTime[i - 1] + interval;
        }
    }

    // Calculate Lagrangian coefficients
    offsetTime = YourTimeDateClass::seconds( epoch - m_segmentStartTime[segment] ) - 
                  interval * (index0 - m_segmentStartIndex[segment]);

    // Outer loop to calculate coefficient for each term.
	for( i = 0; i < m_numInterp; i++ )
	{
		m_lagCoeffs[i] = 1.0;

        // Inner loop to calculate products
		for(long j = 0; j < m_numInterp; j++ )
		{
			if( i != j )
				m_lagCoeffs[i] *= (offsetTime - aTime[j])/( aTime[i] - aTime[j] );
		}
	}
	return;
}


/******************************************************************************
 This routine calculates an interpolated value using the coefficients for
 the Lagrangian interpolation formula. For a time series of xi, the x value 
 using an 8 point interpolator is

  x = c0 * x0 + c1 * x1 + ... + c7 * x7
  
 The ci are the previously calculated coefficients.
 Reference: Eq (3.1.1), P. 102, Numerical Recipes in Fortran, William Press, et al.
******************************************************************************/
double Track::lagrangianInterpolation(const std::vector<double>&aX)
{
    yourASSERTION( m_numInterp > 0 &&  m_numInterp <= (long) m_lagCoeffs.size()  );

    double x = 0;
	for(long i = 0; i < m_numInterp; i++ )
    {
        x += m_lagCoeffs[i] * aX[ m_lagIndex[i] ];
    }
	return x;
}
YourTimeIntervalClass Track::lagrangianInterpolation(const std::vector<YourTimeIntervalClass>&aX)
{
    yourASSERTION( m_numInterp > 0 &&  m_numInterp <= (long) m_lagCoeffs.size()  );

    double x = 0;
	for(long i = 0; i < m_numInterp; i++ )
    {
        x += m_lagCoeffs[i] * YourTimeDateClass::seconds( aX[ m_lagIndex[i]] );
    }
    return YourTimeDateClass::seconds( x );
}


/******************************************************************************
 Routine to save a preset timebias which will be used when calculating position.
 This value is added to the value of current timebias. Sources of this timebias
 include e.g. from ILRS TBF and Drag function data, or from observations from
 other sites. This method saves the value to the last segment loaded.
******************************************************************************/
void Track::setTimeBiasPreset( YourTimeIntervalClass timeBiasStart,
                               YourTimeIntervalClass timeBiasStop )
{
    yourASSERTION( !m_segmentStartTimeBiasPreset.empty() );
    m_segmentStartTimeBiasPreset[m_segmentStartTime.size()-1] = timeBiasStart;
    m_segmentStopTimeBiasPreset[m_segmentStopTime.size()-1]   = timeBiasStop;
    return;
}

/******************************************************************************
 Routine to ensure azimuth and elevation angles are set to the best quadrant
 and are continuous throughout the data set.
******************************************************************************/
void Track::normalizeAngles(double *previousAz, 
                            double *azimuth,
                            double *elevation )
{
    yourASSERTION(previousAz);
    yourASSERTION(azimuth);
    yourASSERTION(elevation);
    double elev = *elevation;
    double azim = *azimuth;

    // Limit elevation to 90 degress.
    if( fabs(elev) > PI_ON_2 )
    {
        // First ensure that it doesn't exceed 180 degrees.
        elev = fmod( elev, TWO_PI);
        if( elev <= PI ) elev += TWO_PI;
        if( elev >  PI ) elev -= TWO_PI;
        
        // If it exceeds 90 degrees, also flip the azimuth
        if( elev > PI_ON_2 )
        {
            elev = PI - elev;
            azim += TWO_PI;
        }
        else if( elev < PI_ON_2 )
        {
            elev = - PI - elev;
            azim += TWO_PI;
        }
    }

    // Correct for 0 to 360 degree azimuth discontinuities.
    if (*previousAz == 0.0)
    {
        // Make initial azimuth in range 0-360 degrees. 
        azim = fmod( azim, TWO_PI);
        if ( azim < 0.0 ) 
            azim += TWO_PI;
        *previousAz = azim;
    }

    double deltaAzimuth = azim - *previousAz;
    if (fabs( deltaAzimuth ) > PI)
        azim = (deltaAzimuth < 0) ? azim + TWO_PI : azim - TWO_PI;
    *previousAz  = azim;
    *azimuth     = azim;
    *elevation = elev;
    return;
}

/******************************************************************************
 Routine to calculate required data at a given epoch.
******************************************************************************/
bool Track::calculatePosition( SiteHandle& site, YourTimeDateClass epoch )
{
    long index = -1;
    bool available = processDataPoint( site, epoch, index );

    // If this point is below the site's horizon, then reset the previousIndex counter
    // for the site. This index is used to eliminate discontinuities in passes.
    if( !site.targetAboveHorizon() )
        site.m_previousIndex = -1;

    return available;
}

/******************************************************************************
 Routine to calculate required data at a given epoch.
******************************************************************************/
bool Track::calculatePosition( SiteHandle& site, long index )
{
    // Generate node data if not already generated. 
    if( m_maxIndex < 0 )
    {
        if( !generateDummyNodes( site ) )
            return 0;
    }
    yourASSERTION( !m_segmentStartIndex.empty() );

    YourTimeDateClass epoch;
    bool available = processDataPoint( site, epoch, index );

    // If this point is below the site's horizon, then reset the previousIndex counter
    // for the site. This index is used to eliminate discontinuities in passes.
    if( !site.targetAboveHorizon() )
        site.m_previousIndex = -1;

    return available;
}

/******************************************************************************
 Routine to determine the target position data at a given epoch or index.
******************************************************************************/
bool Track::getITRFData( SiteHandle& site, YourTimeDateClass epoch, long index )
{
    // If the index is given, then extract data for the given node.
    // Otherwise, use the given epoch and calculate data by interpolation.
    if( index >= 0 )
        return extractNodeData( site, index );

    // Set the actual epoch before adjusting for timebias.
    site.m_epoch = epoch;

    // Correct the epoch for current timebias.
    epoch += timeBias(site, epoch);

    // Calculate the index for access to data values, given the requested time.
    site.m_index = getIndex( epoch, &site.m_previousIndex, &site.m_segment );
    if( site.m_index < 0 )
        return false;

    // Use lagrangian interpolation to generate orbit positions.
    generateLagrangianCoefficients( site.m_index, site.m_segment, epoch );
    return interpolateData( site, epoch );
}


/******************************************************************************
Set start epoch
******************************************************************************/
void Track::setStartEpoch(YourTimeDateClass epoch)
{
    if( epoch < m_periodStart )
        yourTHROW(" Start epoch earlier than prediction period start." );
    if( epoch > m_periodStop )
        yourTHROW(" Start epoch later than prediction period end." );

    m_startEpoch = epoch;

    return;
}
/******************************************************************************
Set stop epoch
******************************************************************************/
void Track::setStopEpoch(YourTimeDateClass epoch)
{
    if( epoch < m_periodStart )
        yourTHROW(" Stop epoch earlier than prediction period start." );
    if( epoch > m_periodStop )
        yourTHROW(" Stop epoch later than prediction period end." );

    m_stopEpoch = epoch;

    return;
}

/******************************************************************************
Input a 3x1 matrix of rectangular coordinates (x,y,z) and convert to 
spherical coordinates i.e.  range, azimuth, elevation,  or
              (geocentric) radius, longitude, latitude. 
rho = sqrt(x*x + y*y)
range = radius = sqrt( x*x + y*y + z*z)
azimuth   = longitude = atan(y/x)
elevation = latitude = atan(z/rho)
******************************************************************************/
void Track::convertToSpherical( YourMatrix<double>& pos, YourMatrix<double>& vel )
{
    YourMatrix<double> x(pos);
    double rhosq = x(0,0)* x(0,0) + x(1,0)* x(1,0);

    pos(0,0) = sqrt( rhosq + x(2,0) * x(2,0) );  // range,    radius
    pos(1,0) = atan2( x(1,0), x(0,0) );          // azimuth,  longitude
    pos(2,0) = atan2( x(2,0), sqrt(rhosq) );     // elevation,latitude

    if( rhosq <= 0.0 )
        return;

    YourMatrix<double> u(vel);
    vel(0,0) = ( x(0,0)*u(0,0) + x(1,0)*u(1,0) + x(2,0)*u(2,0) )/pos(0,0);
    vel(1,0) = ( x(0,0)*u(1,0) - x(1,0)*u(0,0) )/rhosq;
    double rho = sqrt(rhosq);
    vel(2,0) = rho * u(2,0) - x(2,0)*( x(0,0)*u(0,0) + x(1,0)*u(1,0) )/rho;
    vel(2,0) /= pos(0,0)*pos(0,0);

    return;
}

/******************************************************************************
 Input a 3x1 matrix of spherical coordinates (radius, longitude, geocentric latitude)
 and convert to rectangular coordinates (x,y,z).
 If required convert spherical velocitites to rectangular velocities.
******************************************************************************/
void Track::convertToRectangular( YourMatrix<double>& pos, YourMatrix<double>& vel )
{
    YourMatrix<double> m( pos );
    double radius = m[0][0];
    double cosLon = cos( m[1][0] );
    double sinLon = sin( m[1][0] );
    double cosLat = cos( m[2][0] );
    double sinLat = sin( m[2][0] );
    double cc = cosLat * cosLon;
    double cs = cosLat * sinLon;
    double sc = sinLat * cosLon;
    double ss = sinLat * sinLon;
    pos(0,0) = radius * cc;
    pos(1,0) = radius * cs;
    pos(2,0) = radius * sinLat;

    YourMatrix<double> u( vel );
    double rdot = u(0,0);
    vel(0,0) = rdot * cc - radius * ( u(1,0) * cs + u(2,0) * sc );
    vel(1,0) = rdot * cs + radius * ( u(1,0) * cc - u(2,0) * ss );
    vel(2,0) = rdot * sinLat + radius * u(2,0) * cosLat;
    return;
}

/******************************************************************************
Compute the relativistic differential relationship, in seconds, to convert an
interval of proper or atomic time to an interval of coordinate time. 
See Mulholland, 1972 PASP, 84, P.357.
Input epoch and spherical (geocentric) coordinates of telescope.
Output is a scale factor (a bit more than 1).
Author: R. L. Ricklefs / UT CSR
******************************************************************************/
double Track::relativisticScale( YourTimeDateClass epoch, double longitude, double latitude )
{
    double julCent = 1.0 + (epoch.mjd() - MJD2000)/DAYS_PER_CENTURY;
    double bd = 1.0e-04 * julCent * DAYS_PER_CENTURY;

    double ems = 358.475845 + bd * ( 0.985600267e4 -
                              bd * ( 1.12e-05      - 
                              bd * ( 0.7e-07        )));
    ems *= RAD_PER_DEG;
    ems = fmod( ems, TWO_PI);
    double ecs = 0.01675104 - bd * ( 1.1444e-05 + 
                              bd * ( 9.4e-09    ));

    // Obtain eccentric anomaly using the above parameters (cd Brouwer and Clemence p76)
    double cosu = cos( ems + ecs * sin(ems) );
    double ut = TWO_PI * ( epoch.mjd() - floor( epoch.mjd() ));

    double scale = 1.0 - 3.302e-10 * cosu /(1.0 - ecs * cosu) -
                   1.476e-10 * cos( latitude )* cos( ut + longitude );
    return scale;
}
