#include "Track.h"
#include "YourException.h"

/******************************************************************************
 Constructor
******************************************************************************/
TransponderTrack::TransponderTrack( YourTimeDateClass startPredictions,
                                         YourTimeDateClass stopPredictions,
                                         YourTimeIntervalClass stepSize ) : OrbitalTrack( startPredictions, stopPredictions, stepSize )
{
    m_trackType = Track::Type::transponder;
}

/******************************************************************************
 Destructor
******************************************************************************/
TransponderTrack::~TransponderTrack( )
{
}

/******************************************************************************
Method to add a new node i.e. a set of prediction data at a given point to the
time series of such points. It also sets up a number of arrays for data
associated with each prediction segment. One segment corresponds to predictions
generated from a given set of predictions. i.e. from a CPF file.
These arrays include for each segment
1. start and stop time - time segments may overlap
2. start and stop index 
3. if we have return leg predictions, velocity, etc.
******************************************************************************/
void TransponderTrack::appendNode( YourTimeDateClass epoch,
                                        bool newSegment, bool biDirectional, bool cCorrApplied,
                                        const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                                        YourTimeIntervalClass ut1_utc, const std::vector<double>& transponder )
{
    OrbitalTrack::appendNode(epoch, newSegment, biDirectional, cCorrApplied, posVelCor, xPolar, yPolar, ut1_utc);

    if( newSegment )
    {
        m_transponderPulseFrequency.push_back(  transponder[0] );
        m_transponderUtcOffset.push_back(       transponder[1] );
        m_transponderTransmitDelay.push_back(   transponder[2] );
        m_transponderOscillatorDrift.push_back( transponder[3] );
    }

    m_aTransponderOscillatorRelCor.push_back( transponder[4] );
    return;
}

/******************************************************************************
 Routine to interpolate required data at a given specific time.
******************************************************************************/
bool TransponderTrack::interpolateData(SiteHandle& site, YourTimeDateClass epoch ) 
{
    OrbitalTrack::interpolateData( site, epoch );
    site.m_transponderCor = lagrangianInterpolation( m_aTransponderOscillatorRelCor );
    return true;
}

/******************************************************************************
 Routine to interpolate required data at a given specific time.
******************************************************************************/
bool TransponderTrack::extractNodeData(SiteHandle& site, long index ) 

{
    if( index < 0 || index >= long(m_aTransponderOscillatorRelCor.size()) )
        return false;

    if( !OrbitalTrack::extractNodeData( site, index ) )
        return false;

    site.m_transponderCor = m_aTransponderOscillatorRelCor[index];

    return true;
}

/******************************************************************************
 Routine to calculate required x,y,z coordinates given a specific time within 
 the available range of data. 
******************************************************************************/
bool TransponderTrack::processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index ) 
{
    // Account for transponder delay.
    m_targetResponseDelay = YourTimeDateClass::seconds( m_transponderTransmitDelay[site.m_segment] );

    // Convert rectangular coordinates to topocentric coordinates for the given site.
    if( !convertITRFToTopocentric(site, epoch, index ) )
        return false;

    // Add relativistic corrections and delays and em geodesic curvature
    site.m_targetRange += site.m_relCor[0] + site.m_relCor[1] + YourTimeDateClass::picoSeconds(270,0);
    site.m_targetRange *= Track::relativisticScale( epoch, site.m_posSpherical[4], site.m_posSpherical[5] );
    site.m_targetRange += m_targetResponseDelay;

    // If the target is below zero degree horizon, then return false;
    normalizeAngles( &site.m_previousAz, &site.m_targetPosSpher[1], &site.m_targetPosSpher[2] );  

    updateTargetCharacteristics( site );

    return true;
}
