/* ========================================================================
C*
      PROGRAM CPF_PGM
CC
CC NAME       :  CPF_PGM
CC
CC PURPOSE    :  COMPUTE AZIMUTH/ELEVATION/RANGE VALUES FOR SATELLITE PASS
CC               USING CPF GEOCENTRIC SATELLITE POSITIONS
CC
CC               SAMPLE PROGRAM
CC
CC SR CALLED  :  CPF_INTER (CALLING HERMITE, DDREH, DMLMAV)
CC               EPHINITU
CC
CC INCLUDES   :  EPHI.INC, EPHIHDR.INC
CC
CC REMARKS    :  STATION COORDINATES ARE "HARDWIRED" INTO THE PROGRAM
CC
CC               THE PROGRAM DOES NOT HANDLE LEAP SECOND TRANSITIONS!
CC
CC AUTHOR     :  W. GURTNER
CC
CC
CC CREATED    :  12-OCT-2005            LAST MODIFIED :
CC
CC CHANGES    :  29-NOV-2005 (WG) : CENTER OF MASS CORRECTION
CC
CC COPYRIGHT  :  ASTRONOMICAL INSTITUTE
CC      2005      UNIVERSITY OF BERNE
CC                    SWITZERLAND
CC
CC  MODIFIED:
CC              17-Oct-2005 : Converted to c; R. Ricklefs UT/CSR
C*
=========================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../include/ephi.h"
struct eh ephihdr;
struct ei ephi;
int debug= 0;	/* cephinitu debug flag */

void cpf_inter();
void hermite();
void ddreh();
void dmlmav();
FILE   *lfncpf, *lfneph;

main ()
{
      char cpffil[40],fileph[40];

/* Station coordinates (xyz, latitude, longitude east>0) */
      double staxyz[3] = {4331283.557e0,567549.902e0,4633140.353e0};
      double stalat = 46.877229e0;
      double stalon = 7.4652415e0;

      double tstart, tend, dt, tbias;
      double tjul, range, fltime, azi, ele, difazi, difele;
      double pi, tim;
      int    i, mode, iunits, irc, ierr;

      /* Number of values to be used for interpolation (sr hermite) */
      int    nval= 10;

      pi= 4.e0*atan(1.e0);

/* Radians */
      stalat*= pi/180.e0;
      stalon*= pi/180.e0;

/*
* Get Input Options from Input Queue
* ----------------------------------
*/
/* Filename of CPF prediction input file */
      printf ("CPF prediction file? ");
      scanf ("%s", cpffil);
/* Filename of ephemeris output file */
      printf ("Output file? ");
      scanf ("%s", fileph);
/* Start time (Modified Julian Date) */
      printf ("Start time (Modified Julian Date)? ");
      scanf ("%lf", &tstart);
/* End time (MJD) */
      printf ("Stop  time (Modified Julian Date)? ");
      scanf ("%lf", &tend);
/* Increment (seconds) */
      printf ("Time increment (seconds)? ");
      scanf ("%lf", &dt);
/* Time Bias (seconds) */
      printf ("A priori time bias (sec, will be added to epochs)? ");
      scanf ("%lf", &tbias);
/*
C  Read CPF File
C  -------------
C
C  JDI,JDF,JDS,OUTV,NV USED BELOW ARE STORED BY SUBR EPHINITU INTO COMMON
C  (EPHEMERIDES: EPOCH, X,Y,Z, NUMBER OF VECTORS)
C  (SEE INCLUDE FILES)
*/
      if ((lfncpf=fopen(cpffil,"r"))==NULL)  {
        printf ("*** PGM CPF_PGM: Error opening CPF file %s\n",cpffil);
        exit(1); }

      ierr= cephinitu (lfncpf);
      close(lfncpf);
      if (ierr != 0) {
        printf("*** PGM CPFINT: Error reading CPF file\n");
        exit(1);
      }

/*
 * Is this an slr target or llr/tlr? */
      if (ephihdr.ltcorr == 1 || ephihdr.tar_type == 2 || ephihdr.compat != 1)
        {
          printf("PGM CPF_PGM: File %c\n", cpffil);
          printf("             requires LLR/TLR prediction software!");
          exit(1);
        }
/*
 *  Topocentric vectors for all epochs, flight time applied
 *  *******************************************************
 */
/* Outbound vector, position of satellite at bounce time
*  azimuth and elevation for laser beam at transmit time
*  azimuth and elevation for receiver at transmit time: add difazi/difele */
      mode= 1;
/* Time units (for jds): seconds */
      iunits=0;

/* File to store the interpolated values */
    if ((lfneph=fopen(fileph,"w"))==NULL)  {
        printf ("Could not open file %s\n",fileph);
        exit(1); }

/* Test modified Julian date */
      i= 0;
      while ((tjul=tstart+i*dt/86400.e0) <= tend+ 1.e-9)
        {
/*        Relative time to start of cpf file 
 *        JDI (FULL DAYS), JDI (FRACTION) GIVEN IN JULIAN DATE
 *        JDS: DAYS RELATIVE TO FIRST EPOCH */
          tim= (tjul-(ephi.jdi[0]+ephi.jdf[0]-2400000.5e0))*86400.e0+tbias;

          cpf_inter(tim,iunits,mode,staxyz,stalat,stalon,ephihdr.cofm_corr,
                   ephi.jds,&ephi.outv[0][0],&ephi.outv[1][0],&ephi.outv[2][0],
                   ephi.nv,nval, &range,&fltime,&azi,&ele,&difazi,&difele,&irc);
	  fprintf (lfneph,"%11.5f  %10.5f  %10.5f  %8.5f  %8.5f  %13.10f\n",
                   tjul,azi,ele,difazi,difele,fltime);
          i++;
        }
      close(lfneph);
      exit(0);

}
