C*
      SUBROUTINE CPF_INTER(TT,IUNITS,MODE,STAGEO,STALAT,STALON,
     1                     COFM_CORR,TTAB,OBJTAB_X,OBJTAB_Y,OBJTAB_Z,
     2                     NTAB,NINT,RANGE,FLTIME,AZIOUT,ELEOUT,
     3                     DIFAZI,DIFELE,IRCODE)
CC
CC  PURPOSE   : INTERPOLATE SATELLITE POSITION, DETERMINE PREDICTION
CC              ELEMENTS
CC
CC  PARAMETERS:
CC         IN : TT       : EPOCH (TRANSMIT=FIRING TIME)            R*8
CC              IUNITS   : UNITS OF TIME (0 = SECONDS, #0 = DAYS)  I*4
CC                         (FOR TT AND TTAB!)
CC              MODE     : 0: NO FLIGHT TIME APPLIED               I*4
CC                            FLIGHT TIME = 2 * INSTANT RANGE
CC                                          AT EPOCH TT
CC                         1: OUTBOUND VECTOR (AZIMUTH, ELEVATION)
CC                            FLIGHT TIME = 2 * INSTANT RANGE AT
CC                                          BOUNCE TIME
CC                         2: OUTBOUND AND INBOUND VECTOR
CC                            FLIGHT TIME = OUTBOUND + INBOUND
CC              STAGEO   : GEOCENTRIC STATION COORDINATES XYZ (M)  R*8(*)
CC              STALON   : STATION LONGITUDE (EAST  > 0, RADIANS)  R*8
CC              STALAT   : STATION LATITUDE  (NORTH > 0, RADIANS)  R*8
CC              COFM_CORR: RADIAL CENTER OF MASS CORRECTION (M)    R*8
CC              TTAB     : TABULATED EPOCHS (SECONDS OR DAYS,      R*8
CC                                           SEE IUNITS)
CC              OBJTAB_X : TABULATED X-COORD. OF OBJECT (M)        R*8(*)
CC              OBJTAB_Y : TABULATED Y-COORD. OF OBJECT (M)        R*8(*)
CC              OBJTAB_Z : TABULATED Z-COORD. OF OBJECT (M)        R*8(*)
CC              NTAB     : NUMBER OF TABULATED EPOCHS/COORDINATES  I*4
CC              NINT     : NUMBER OF TABULATED VALUES TO USE       I*4
CC                         (ARGUMENT FOR LAGRANGE INTERPOLATION)
CC        OUT : RANGE    : ONE-WAY RANGE (M)                       R*8
CC              FLTIME   : TWO-WAY FLIGHT TIME (SEC)               R*8
CC              AZIOUT   : OUTBOUND AZIMUTH   (DEG)                R*8
CC              ELEOUT   : OUTBOUND ELEVATION (DEG)                R*8
CC              DIFAZI   : POINT-BEHIND (AZIMUTH,   DEGREES)       R*8
CC              DIFELE   : POINT-BEHIND (ELEVATION, DEGREES)       R*8
CC                         DIFFERENCE INBOUND MINUS OUTBOUND AT
CC                         EPOCH TT
CC              IRCODE   : RETURN CODE 0 = OK                      I*4
CC                                     1 = INTERPOLATION NOT IN
CC                                         CENTER INTERVAL
CC                                     2 = ERROR IN INTERPOLATION
CC
CC  CALLS:      HERMITE, DDREH, DMLMAV
CC
CC  AUTHOR:     W. GURTNER
CC              ASTRONOMICAL INSTITUTE
CC              UNIVERSITY OF BERNE
CC
CC  CREATED:    25-JUL-2005
CC
CC  MODIFIED:    2-AUG-2005 : TIME UNITS (SEC OR DAYS), MODE
CC              29-NOV-2005 : RADIAL CENTER OF MASS CORRECTION
CC              30-JUN-2009 : Comments: DIFAZI and DIFELE are degrees not asec.
C*
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 STAGEO(3)
      REAL*8 TTAB(*),OBJTAB_X(*),OBJTAB_Y(*),OBJTAB_Z(*)
      REAL*8 OBJBNC(3),OBJTRT(3),TOPINB(3),TOPOUT(3),TOPTRT(3)
      REAL*8 R2LA(3,3),R3PI(3,3),R3LO(3,3),R3SD(3,3),R3(3,3),STAROT(3)
      DATA IFIRST/1/
      SAVE
C
      IF (IFIRST.EQ.1) THEN
C
C  SPEED OF LIGHT
        C=299792458.D0
C
        PI=4.D0*DATAN(1.D0)
        PRVLON=999.D0
        PRVLAT=999.D0
        IFIRST=0
      END IF
C
C  DDREH:  Computation of rotation matrices
C  R3LO AROUND LONGITUDE, R2LA AROUND PI/2-LATITUDE, R3PI AROUND PI
      IF(STALON.NE.PRVLON.OR.STALAT.NE.PRVLAT) THEN
        CALL DDREH(3,STALON,R3LO)
        CALL DDREH(2,PI/2-STALAT,R2LA)
        CALL DDREH(3,PI,R3PI)
        PRVLON=STALON
        PRVLAT=STALAT
      END IF
C
C  INSTANT VECTOR AT TRANSMIT TIME
C  *******************************
C
C  INTERPOLATE GEOCENTRIC POSITION OF THE OBJECT FOR TRANSMIT TIME
      CALL HERMITE(1,TTAB,OBJTAB_X,Z,NTAB,NINT,TT,OBJTRT(1),VX,IRCX)
      CALL HERMITE(1,TTAB,OBJTAB_Y,Z,NTAB,NINT,TT,OBJTRT(2),VY,IRCY)
      CALL HERMITE(1,TTAB,OBJTAB_Z,Z,NTAB,NINT,TT,OBJTRT(3),VZ,IRCZ)
      IF(IRCX+IRCY+IRCZ.NE.0) THEN
        WRITE(*,1) IRCX,IRCY,IRCZ
1       FORMAT(/,' SR CPF_INTER: Return codes from SR HERMITE:',3I3)
        IRCODE=2
        RETURN
      ENDIF
C
C  TOPOCENTRIC VECTOR STATION/OBJECT BOTH AT TRANSMIT TIME
      TOPTRT(1)=OBJTRT(1)-STAGEO(1)
      TOPTRT(2)=OBJTRT(2)-STAGEO(2)
      TOPTRT(3)=OBJTRT(3)-STAGEO(3)
C
C  INSTANT DISTANCE FROM STATION TO OBJECT AT TRANSMIT TIME
      DISTTRT=DSQRT(TOPTRT(1)**2+TOPTRT(2)**2+TOPTRT(3)**2)
C
C  TOPOCENTRIC VECTOR IN LOCAL SYSTEM
C
C  DMLMAV: MULTIPLY ROTATION MATRIX WITH VECTOR
      CALL DMLMAV(TOPTRT,R3LO,TOPTRT)
      CALL DMLMAV(TOPTRT,R2LA,TOPTRT)
      CALL DMLMAV(TOPTRT,R3PI,TOPTRT)
C
C  AZIMUTH AND ELEVATION (degrees)
      ELETRT=DATAN(TOPTRT(3)/DSQRT(TOPTRT(1)**2+TOPTRT(2)**2))*180/PI
      AZITRT=DATAN2(-TOPTRT(2),TOPTRT(1))*180/PI
      IF(AZITRT.LT.0.D0) AZITRT=AZITRT+360
C
C  INSTANT VECTOR ONLY
      IF(MODE.EQ.0) THEN
C  ONE-WAY RANGE
        RANGE=DISTTRT-COFM_CORR
C  ROUND TRIP FLIGHT TIME (SEC)
        FLTIME=2*RANGE/C
C  DIRECTION
        AZIOUT=AZITRT
        ELEOUT=ELETRT
C  NO POINT BEHIND
        DIFAZI=0
        DIFELE=0
        RETURN
      END IF
C
C  OUTBOUND VECTOR
C  ***************
C
C  INITIALIZE
      STAROT(1)=STAGEO(1)
      STAROT(2)=STAGEO(2)
      STAROT(3)=STAGEO(3)
      TOUT=DISTTRT/C
C
C  ITERATION
C
      DO 30 I=1,2
C
C  BOUNCE TIME
        IF(IUNITS.EQ.0) THEN
          TB=TT+TOUT
        ELSE
          TB=TT+TOUT/86400.D0
        END IF
C
C  INTERPOLATE GEOCENTRIC POSITION OF THE OBJECT FOR BOUNCE TIME TB
        CALL HERMITE(1,TTAB,OBJTAB_X,Z,NTAB,NINT,TB,OBJBNC(1),VX,IRCX)
        CALL HERMITE(1,TTAB,OBJTAB_Y,Z,NTAB,NINT,TB,OBJBNC(2),VY,IRCY)
        CALL HERMITE(1,TTAB,OBJTAB_Z,Z,NTAB,NINT,TB,OBJBNC(3),VZ,IRCZ)
        IF(IRCX+IRCY+IRCZ.NE.0) THEN
          WRITE(*,1) IRCX,IRCY,IRCZ
          IRCODE=2
          RETURN
        ENDIF
C       
C  TOPOCENTRIC OUTBOUND VECTOR
        TOPOUT(1)=OBJBNC(1)-STAROT(1)
        TOPOUT(2)=OBJBNC(2)-STAROT(2)
        TOPOUT(3)=OBJBNC(3)-STAROT(3)
C
C  DISTANCE FROM STATION (TT) TO OBJECT (TB)
        DISTOUT=DSQRT(TOPOUT(1)**2+TOPOUT(2)**2+TOPOUT(3)**2)
C
C  OUTBOUND FLIGHT TIME (SEC)
        TOUT=DISTOUT/C
C
C  ROTATE STATION DURING FLIGHT TIME (RADIANS)
        DSIDT=6.300388D0*(TOUT/86400.D0)
        CALL DDREH(3,DSIDT,R3SD)
        CALL DMLMAV(STAGEO,R3SD,STAROT)
C
30    CONTINUE
C
C  TOPOCENTRIC OUTBOUND VECTOR IN LOCAL SYSTEM
C
      CALL DMLMAV(TOPOUT,R3LO,TOPOUT)
      CALL DMLMAV(TOPOUT,R2LA,TOPOUT)
      CALL DMLMAV(TOPOUT,R3PI,TOPOUT)
C
C  OUTBOUND AZIMUTH AND ELEVATION (LASER BEAM POINTING DIRECTION)
C  **************************************************************
C
      ELEOUT=DATAN(TOPOUT(3)/DSQRT(TOPOUT(1)**2+TOPOUT(2)**2))*180/PI
      AZIOUT=DATAN2(-TOPOUT(2),TOPOUT(1))*180/PI
      IF(AZIOUT.LT.0.D0) AZIOUT=AZIOUT+360
CCC   WRITE(*,14) AZIOUT,ELEOUT
14    FORMAT(' Azi/Ele outbound vector : ',2f10.4)
C
C  TOTAL POINT AHEAD / BEHIND
C  **************************
C  DIFFERENCE BETWEEN RECEIVE AND TRANSMIT DIRECTION AT TRANSMIT TIME
C
      DIFAZI=2*(AZITRT-AZIOUT)
      IF(DIFAZI.LT.-360) DIFAZI=DIFAZI+720
      IF(DIFAZI.GT.+360) DIFAZI=DIFAZI-720
      DIFELE=2*(ELETRT-ELEOUT)
CCC   WRITE(*,13) DIFAZI*3600,DIFELE*3600
13    FORMAT(' Point-behind (AZI,ELE arcsec) : ',2F8.1)
C
      IF(MODE.EQ.1) THEN
C
C  AVERAGE DISTANCE FROM STATION TO OBJECT
C  (BOTH AT BOUNCE TIME: GOOD APPROXIMATION)
C  *****************************************
        RANGE=DSQRT((OBJBNC(1)-STAGEO(1))**2+
     1              (OBJBNC(2)-STAGEO(2))**2+
     2              (OBJBNC(3)-STAGEO(3))**2)
C
C  RADIAL CENTER OF MASS CORRECTION
	RANGE=RANGE-COFM_CORR
C
C  ROUND TRIP FLIGHT TIME (SEC)
C  **********************
        FLTIME=2*RANGE/C
C
        RETURN
C
      END IF
C
      IF(MODE.EQ.2) THEN
C
C  INBOUND VECTOR: STATION AT RECEIVING TIME
C  *****************************************
C
        TINB=TOUT
C
C  ITERATION
        DO 40 I=1,2
C
C  RECEIVE TIME
          IF(IUNITS.EQ.0) THEN
            TR=TB+TINB
          ELSE
            TR=TB+TINB/86400.D0
          END IF
C
C  ROTATE STATION DURING INBOUND FLIGHT TIME (OTHER DIRECTION)
          DSIDT=6.300388D0*(TINB/86400.D0)
          CALL DDREH(3,-DSIDT,R3)
          CALL DMLMAV(STAGEO,R3,STAROT)
C
C  TOPOCENTRIC INBOUND VECTOR IN GEOCENTRIC SYSTEM
          TOPINB(1)=OBJBNC(1)-STAROT(1)
          TOPINB(2)=OBJBNC(2)-STAROT(2)
          TOPINB(3)=OBJBNC(3)-STAROT(3)
C
C  DISTANCE FROM OBJECT TO STATION
          DISTINB=DSQRT(TOPINB(1)**2+TOPINB(2)**2+TOPINB(3)**2)
C
C  INBOUND FLIGHT TIME (SEC)
          TINB=DISTINB/C
C
40      CONTINUE
C
C  TOTAL DISTANCE FROM STATION TO OBJECT TO STATION (CORRECT COMPUTATION)
C
        DISTOT=DISTOUT+DISTINB
        FLTIME=(DISTOT-2.D0*COFM_CORR)/C
        RANGE = DISTOT/2.D0-COFM_CORR
      END IF
C
      RETURN
      END
